import tensorflow as tf
import gdown
import os
from .yoloLoss import yolo_loss_tf



# load model
def load_model(model_path = None):

    # if model path is not given
    if model_path is None:
        raise ValueError("model_path is None, please provide a valid path or try \"YOLO\"")
    
    elif model_path == "YOLO_coco":
        download_model_from_gdrive("https://drive.google.com/file/d/1HuiUq4q1mJdlX9PKmGnL505YPTHE8DLZ/view?usp=sharing")
        model_path = 'yolo_tf.h5'


    model = tf.keras.models.load_model(model_path, custom_objects={"yolo_loss_tf": yolo_loss_tf})
    return model





def download_model_from_gdrive(url, destination='yolo_tf.h5'):
    dest = os.path.abspath(destination)

    # Skip if the file already exists and is not empty
    if os.path.isfile(dest) and os.path.getsize(dest) > 0:
        print(f"[gdown] Found existing file → {dest}  (skipping download)")
        return dest

    print("[gdown] Downloading model …")
    gdown.download(url, dest, quiet=False, fuzzy=True)  # resumes if partial
    print(f"[gdown] Model saved to {dest}")
    return dest







if __name__ == '__main__':



    # model_path = '20250215_2106/yolo_model_epoch_63.h5'
    # model = load_model(model_path)
    # print(model.summary())

    # model_path = '20250215_2106/yolo_tf.h5'
    # model = load_model(model_path)
    # print(model.summary())

    # model_path = './yolo_tf.h5'
    # model = load_model(model_path)
    # print(model.summary())


    model = load_model('YOLO')
    print(model.summary())


import random
import numpy as np
import cv2

def random_horizontal_flip(image, bboxes, prob=0.5):
    """
    :param image:  NumPy array (H x W x 3), dtype float or uint8
    :param bboxes: List of bounding boxes, each in [cls, x_center, y_center, w, h] (YOLO format)
    :param flip_prob: Probability of flipping
    :return: (image, bboxes) after potential horizontal flip
    """
    if random.random() < prob:
        # Flip image horizontally
        image = cv2.flip(image, 1)
        
        # Flip bboxes: x_center -> 1 - x_center
        # y_center, w, h stay the same
        # cls stays the same
        new_bboxes = []

                # if no bounding boxes, return
        if bboxes == None:
            return image, bboxes
        
        for box in bboxes:
            cls_id, x, y, w, h = box
            x_new = 1.0 - x
            new_bboxes.append([cls_id, x_new, y, w, h])
        bboxes = new_bboxes

    return image, bboxes


def random_color_jitter(image, brightness_range, contrast_range, prob):
    """
    Randomly adjusts brightness and contrast of the image.
    :param image: NumPy array in range [0,1] or [0,255]
    :param brightness_range: float, max fraction for brightness shift
    :param contrast_range: float, max fraction for contrast shift
    :param p: probability to apply this augmentation
    :return: augmented image
    """
    if random.random() < prob:
        # Convert to float32 for robust operations
        img_float = image.astype(np.float32)        

        # Random brightness
        brightness_factor = random.uniform(-brightness_range, brightness_range)
        # Random contrast
        contrast_factor = random.uniform(-contrast_range, contrast_range)

        # Apply adjustments
        img_float = img_float + (img_float-127.5) * contrast_factor + brightness_factor * 255.0

        # Clip to valid range
        img_float = np.clip(img_float, 0, 255)
        
        
        # Convert back to original dtype
        image = img_float.astype(image.dtype)

    return image


def random_crop(image, bboxes, scale_range=(0.8, 1.0), prob=0.5):
    """
    Randomly crops a region from the image, updates YOLO-format bounding boxes accordingly.
    
    :param image:   NumPy array (H x W x 3), BGR or RGB, in [0,255] or [0,1].
    :param bboxes:  List of bounding boxes in YOLO format: [cls, x_center, y_center, w, h],
                    with x_center, y_center, w, h normalized to [0, 1].
    :param scale_range:  (min_scale, max_scale) for the crop dimension as a fraction of the original size.
    :param p:       Probability of applying the crop.
    :return:        (cropped_image, new_bboxes)
    """
    if random.random() > prob:
        # No crop
        return image, bboxes

    h, w = image.shape[:2]
    
    # Choose random crop size
    min_scale, max_scale = scale_range
    crop_scale = random.uniform(min_scale, max_scale)
    crop_h = int(h * crop_scale)
    crop_w = int(w * crop_scale)

    # Choose top-left corner for the crop
    # We can place the crop anywhere in the original image
    y0 = random.randint(0, h - crop_h)
    x0 = random.randint(0, w - crop_w)

    # Crop the image
    cropped_image = image[y0:y0+crop_h, x0:x0+crop_w]
    resized_image = cv2.resize(cropped_image, (w, h))

    # Adjust bounding boxes
    new_bboxes = []

    # if no bounding boxes, return
    if bboxes == None:
        return resized_image, bboxes
    for box in bboxes:
        cls_id, x_rel, y_rel, w_rel, h_rel = box
        x_center_abs = x_rel * w
        y_center_abs = y_rel * h
        bw_abs = w_rel * w
        bh_abs = h_rel * h
        
        # Convert to [x1, y1, x2, y2] in absolute coords
        x1 = x_center_abs - bw_abs/2
        y1 = y_center_abs - bh_abs/2
        x2 = x_center_abs + bw_abs/2
        y2 = y_center_abs + bh_abs/2
            
        # Clip coords to the crop region
        # Shift them by (x0, y0)
        x1_c = x1 - x0 # if x1 is outside the crop, it will be negative
        y1_c = y1 - y0 # if y1 is outside the crop, it will be negative
        x2_c = x2 - x0 # if x2 is outside the crop, it will be larger than crop_w
        y2_c = y2 - y0 # if y2 is outside the crop, it will be larger than crop_h
        
        # If the box is completely outside the crop, skip it
        # or you could partially clip it (below).
        # We'll partially clip:
        x1_c = max(0, x1_c)
        y1_c = max(0, y1_c)
        x2_c = min(crop_w, x2_c)
        y2_c = min(crop_h, y2_c)
        
        new_w = x2_c - x1_c
        new_h = y2_c - y1_c
        
        # Discard if no area
        if new_w <= 1 or new_h <= 1:
            continue
        
        # Convert back to YOLO [x_c, y_c, w, h] in normalized coords
        x_c_rel = (x1_c + new_w/2) / crop_w
        y_c_rel = (y1_c + new_h/2) / crop_h
        w_rel = new_w / crop_w
        h_rel = new_h / crop_h
        
        # Ensure coords are within [0, 1]
        x_c_rel = min(max(x_c_rel, 0), 1)
        y_c_rel = min(max(y_c_rel, 0), 1)
        w_rel = min(w_rel, 1)
        h_rel = min(h_rel, 1)
        
        new_bboxes.append([cls_id, x_c_rel, y_c_rel, w_rel, h_rel])
        # print('cls', cls_id)
        # print('x1_c, y1_c, x2_c, y2_c:', x1_c, y1_c, x2_c, y2_c)


    

    return resized_image, new_bboxes



def letterbox_and_pad(image, bboxes, final_size=320, color=(114, 114, 114), prob = 0.5):
    
    orig_h, orig_w = image.shape[:2]

    aspect_ratio = orig_w / orig_h

    if aspect_ratio < 1.6 and aspect_ratio > 0.6:
        if random.random() < prob:
            # resize image to make it square
            image = cv2.resize(image, (final_size, final_size))
            return image, bboxes
    


    # Calculate scale to fit the image in a square of size final_size
    scale = min(final_size / orig_w, final_size / orig_h)
    new_w = int(orig_w * scale)
    new_h = int(orig_h * scale)

    # Resize the image
    resized = cv2.resize(image, (new_w, new_h), interpolation=cv2.INTER_LINEAR)

    # Create the padded canvas
    # final_size x final_size, fill with "color"
    padded_image = np.full((final_size, final_size, 3), color, dtype=resized.dtype)

    # Compute top-left corner to place the resized image (we'll center it)
    dw = (final_size - new_w) // 2
    dh = (final_size - new_h) // 2

    # Place resized image onto padded canvas
    padded_image[dh:dh+new_h, dw:dw+new_w] = resized

    # Update bounding boxes: scale + shift
    new_bboxes = []

    # if no bounding boxes, return
    if bboxes == None:
        return padded_image, bboxes
    
    for box in bboxes:
        # Convert YOLO -> absolute
        cls_id, x_rel, y_rel, w_rel, h_rel = box
        x_center_abs = x_rel * orig_w
        y_center_abs = y_rel * orig_h
        w_abs = w_rel * orig_w
        h_abs = h_rel * orig_h
        x1 = x_center_abs - w_abs / 2
        x2 = x_center_abs + w_abs / 2
        y1 = y_center_abs - h_abs / 2
        y2 = y_center_abs + h_abs / 2

        # Scale coords
        x1s = x1 * scale
        x2s = x2 * scale
        y1s = y1 * scale
        y2s = y2 * scale
        # Shift by (dw, dh)
        x1p = x1s + dw
        x2p = x2s + dw
        y1p = y1s + dh
        y2p = y2s + dh

        # Ensure coords are within [0..final_size]
        x1p = max(0, min(final_size, x1p))
        x2p = max(0, min(final_size, x2p))
        y1p = max(0, min(final_size, y1p))
        y2p = max(0, min(final_size, y2p))

        bw = x2p - x1p
        bh = y2p - y1p
        if bw < 1 or bh < 1:
            # Discard boxes too small after scaling/padding
            continue

        # Convert back to YOLO normalized
        x_c = (x1p + x2p) / 2.0
        y_c = (y1p + y2p) / 2.0
        w_box = bw
        h_box = bh
        x_rel = x_c / final_size
        y_rel = y_c / final_size
        w_rel = w_box / final_size
        h_rel = h_box / final_size

        new_bboxes.append([cls_id, x_rel, y_rel, w_rel, h_rel])

    return padded_image, new_bboxes


import cv2
import numpy as np
import os

from . import labelDecode
from .drawImage import draw_result


def predict(model, frame, conf=0.5):
    img_size = 320
    high_stride = model.output_shape[0][1]
    low_stride = model.output_shape[1][1]
    
    frame = cv2.resize(frame, (img_size, img_size))

    result = model.predict(frame[np.newaxis, ...])
    annotations = labelDecode.decode(result[0][0], result[1][0], high_stride, low_stride, conf)
    return annotations



def predict_and_show(model, frame, conf=0.5, cls_names=None):
    img_size = 320
    high_stride = model.output_shape[0][1]
    low_stride = model.output_shape[1][1]
    
    frame = cv2.resize(frame, (img_size, img_size))

    result = model.predict(frame[np.newaxis, ...])
    annotations = labelDecode.decode(result[0][0], result[1][0], high_stride, low_stride, conf)
    result_image = draw_result(np.array(image), annotations, cls_names)
    cv2.imshow('image', result_image)
    cv2.waitKey(0)
    return annotations








# test code
if __name__ == '__main__':
    from load_model import load_model

    data_path = 'C:/Users/osy04/Desktop/wok_me/project/yolo/images/train2017/'
    label_path = 'C:/Users/osy04/Desktop/wok_me/project/yolo/labels_bbox/train2017/'

    file_img = os.listdir(data_path)
    file_txt = os.listdir(label_path)

    model = load_model("YOLO")
    

    # # method1: load model and predict
    # for i in range(5):
    #     image = cv2.imread(data_path + file_img[i])/255.0
    #     result_annotation = predict(model, image)
    #     result_image = draw_result(np.array(image), result_annotation)
    #     cv2.imshow('image', result_image)
    #     cv2.waitKey(0)


    # coco dataset cls names
    coco_cls_names = ['person', 'bicycle', 'car', 'motorcycle', 'airplane', 'bus', 'train', 'truck', 'boat',
                    'traffic light', 'fire hydrant', 'stop sign', 'parking meter', 'bench', 'bird', 'cat', 'dog',
                    'horse', 'sheep', 'cow', 'elephant', 'bear', 'zebra', 'giraffe', 'backpack', 'umbrella',
                    'handbag', 'tie', 'suitcase', 'frisbee', 'skis', 'snowboard', 'sports ball', 'kite', 'baseball bat',
                    'baseball glove', 'skateboard', 'surfboard', 'tennis racket', 'bottle', 'wine glass', 'cup', 'fork',
                    'knife', 'spoon', 'bowl', 'banana', 'apple', 'sandwich', 'orange', 'broccoli', 'carrot', 'hot dog',
                    'pizza', 'donut', 'cake', 'chair', 'couch', 'potted plant', 'bed', 'dining table', 'toilet', 'tv',
                    'laptop', 'mouse', 'remote', 'keyboard', 'cell phone', 'microwave', 'oven', 'toaster', 'sink',
                    'refrigerator', 'book', 'clock', 'vase', 'scissors', 'teddy bear', 'hair drier', 'toothbrush']
    

    # method2: load model and predict
    for i in range(5):
        image = cv2.imread(data_path + file_img[i])/255.0
        result = predict_and_show(model, image, cls_names=coco_cls_names)



from oneclickai.YOLO import fit_yolo_model


#% encode data
# encode training data
train_data_path = 'C:/Users/osy04/Desktop/image'
train_label_path = 'C:/Users/osy04/Desktop/image'

# # encode validation data
val_data_path = 'C:/Users/osy04/Desktop/image'
val_label_path = 'C:/Users/osy04/Desktop/image'

# fit model
fit_yolo_model(train_data_path, train_label_path, val_data_path, val_label_path, epochs=10)


import cv2
import numpy as np
from .drawImage import draw_result
from .predict import predict

def stream(model, cls_names=None):

    capture = cv2.VideoCapture(0)
    if not capture.isOpened():
        print("Error: Could not open video.")
        return



    while True:
        _, frame = capture.read()
        frame = frame/255.0

        annotations = predict(model, frame)
        disp_image = draw_result(np.array(frame), annotations, cls_names)

        cv2.imshow('frame', disp_image)
        if cv2.waitKey(1) & 0xFF == ord('q'):
            break

    capture.release()
    cv2.destroyAllWindows()


# example usage
if __name__ == '__main__':
    from load_model import load_model
    model = load_model("YOLO")

    # coco dataset cls names
    coco_cls_names = ['person', 'bicycle', 'car', 'motorcycle', 'airplane', 'bus', 'train', 'truck', 'boat',
                    'traffic light', 'fire hydrant', 'stop sign', 'parking meter', 'bench', 'bird', 'cat', 'dog',
                    'horse', 'sheep', 'cow', 'elephant', 'bear', 'zebra', 'giraffe', 'backpack', 'umbrella',
                    'handbag', 'tie', 'suitcase', 'frisbee', 'skis', 'snowboard', 'sports ball', 'kite', 'baseball bat',
                    'baseball glove', 'skateboard', 'surfboard', 'tennis racket', 'bottle', 'wine glass', 'cup', 'fork',
                    'knife', 'spoon', 'bowl', 'banana', 'apple', 'sandwich', 'orange', 'broccoli', 'carrot', 'hot dog',
                    'pizza', 'donut', 'cake', 'chair', 'couch', 'potted plant', 'bed', 'dining table', 'toilet', 'tv',
                    'laptop', 'mouse', 'remote', 'keyboard', 'cell phone', 'microwave', 'oven', 'toaster', 'sink',
                    'refrigerator', 'book', 'clock', 'vase', 'scissors', 'teddy bear', 'hair drier', 'toothbrush']

    stream(model, coco_cls_names)


#%%
import numpy as np
import os
import cv2
import tensorflow as tf

from .readTxt import load_data_from_txt
from .labelEncode import encode
from .createDataAugumentation import random_horizontal_flip, random_color_jitter, random_crop, letterbox_and_pad


def data_generator(num_classes, high_stride, low_stride, img_size,
                   data_path, label_path, batch_size=32, shuffle=True, prob=0.5):

    # List image files (without extension)
    image_files = [f[:-4] for f in os.listdir(data_path) if f.endswith('.jpg')]
    num_images = len(image_files)
    indices = np.arange(num_images)

    # Preallocate label shapes
    high_label_shape = (high_stride, high_stride, num_classes + 5)
    low_label_shape = (low_stride, low_stride, num_classes + 5)

    # shuffle
    if shuffle:
        np.random.shuffle(indices)

    for start in range(0, num_images, batch_size):
        end = min(start + batch_size, num_images)
        batch_indices = indices[start:end]
        current_batch_size = len(batch_indices)
        
        # Preallocate arrays for the batch
        x_batch = np.empty((current_batch_size, img_size, img_size, 3), dtype=np.float32)
        y_batch_high = np.empty((current_batch_size, *high_label_shape), dtype=np.float32)
        y_batch_low = np.empty((current_batch_size, *low_label_shape), dtype=np.float32)
        
        for i, idx in enumerate(batch_indices):
            image_id = image_files[idx]
            # Load and process the image
            image_path = os.path.normpath(os.path.join(data_path, image_id + '.jpg'))
            image = cv2.imread(image_path)

            # Load annotation
            label_file = os.path.join(label_path, image_id + '.txt')
            annotation = load_data_from_txt(label_file)

            # image resize
            # image = cv2.resize(image, (img_size, img_size))

            # pad if necessary with a probabiliy
            image, annotation = letterbox_and_pad(image, annotation, final_size=img_size, prob=prob)

            # Data Augmentation
            image, annotation = random_horizontal_flip(image, annotation, prob=prob)
            image = random_color_jitter(image, brightness_range=0.2, contrast_range=0.4, prob=prob)
            image, annotation = random_crop(image, annotation, scale_range=(0.5, 1.0), prob=prob)

            # Encode labels
            high_label, low_label = encode(
                annotation, num_classes, high_stride, low_stride, area_threshold=0.08
            )
            
            image = image / 255.0  # Normalize
            x_batch[i] = image
            y_batch_high[i] = high_label
            y_batch_low[i] = low_label
        

        yield x_batch, (y_batch_high, y_batch_low)


# Wrap the generator in a tf.data.Dataset:
def create_tf_dataset(num_classes, high_stride, low_stride, img_size,
                      data_path, label_path, batch_size=32, shuffle=True, prob=0.5):
    output_signature = (
        tf.TensorSpec(shape=(None, img_size, img_size, 3), dtype=tf.float32),
        (
            tf.TensorSpec(shape=(None, high_stride, high_stride, num_classes+5), dtype=tf.float32),
            tf.TensorSpec(shape=(None, low_stride, low_stride, num_classes+5), dtype=tf.float32)
        )
    )
    
    dataset = tf.data.Dataset.from_generator(
        lambda: data_generator(num_classes, high_stride, low_stride, img_size,
                                 data_path, label_path, batch_size, shuffle, prob),
        output_signature=output_signature
    )
    # Prefetch allows the next batch to be prepared while the current one is being processed.
    dataset = dataset.prefetch(buffer_size=tf.data.AUTOTUNE)
    return dataset


#%%
# Example usage:
if __name__ == '__main__':
    from drawImage import draw_result_imshow
    import labelDecode
    data_path = 'C:/Users/osy04/Desktop/wok_me/project/yolo/images/val2017'
    label_path = 'C:/Users/osy04/Desktop/wok_me/project/yolo/labels_bbox/val2017'
    num_classes = 80
    high_stride = 20
    low_stride = 5
    img_size = 360
    batch_size = 32
    
    # Create the generator
    train_gen = data_generator(num_classes, high_stride, low_stride, img_size,
                               data_path, label_path, batch_size=1, shuffle=True)
    
    x_batch, [y_batch_high, y_batch_low] = next(train_gen)
    print("x_batch shape:", x_batch.shape)
    print("y_batch_high shape:", y_batch_high.shape)
    print("y_batch_low shape:", y_batch_low.shape)

    # Create the tf.data.Dataset
    train_dataset = create_tf_dataset(num_classes, high_stride, low_stride, img_size,
                                      data_path, label_path, batch_size=batch_size, shuffle=True)
    

    coco_cls_names = ['person', 'bicycle', 'car', 'motorcycle', 'airplane', 'bus', 'train', 'truck', 'boat',
                'traffic light', 'fire hydrant', 'stop sign', 'parking meter', 'bench', 'bird', 'cat', 'dog',
                'horse', 'sheep', 'cow', 'elephant', 'bear', 'zebra', 'giraffe', 'backpack', 'umbrella',
                'handbag', 'tie', 'suitcase', 'frisbee', 'skis', 'snowboard', 'sports ball', 'kite', 'baseball bat',
                'baseball glove', 'skateboard', 'surfboard', 'tennis racket', 'bottle', 'wine glass', 'cup', 'fork',
                'knife', 'spoon', 'bowl', 'banana', 'apple', 'sandwich', 'orange', 'broccoli', 'carrot', 'hot dog',
                'pizza', 'donut', 'cake', 'chair', 'couch', 'potted plant', 'bed', 'dining table', 'toilet', 'tv',
                'laptop', 'mouse', 'remote', 'keyboard', 'cell phone', 'microwave', 'oven', 'toaster', 'sink',
                'refrigerator', 'book', 'clock', 'vase', 'scissors', 'teddy bear', 'hair drier', 'toothbrush']
    
    for i in range(10):
        x_data, [y_data_high, y_data_low] = next(iter(train_gen))
        print(x_data.shape, y_data_high.shape, y_data_low.shape)
        box = labelDecode.decode(y_data_high[0], y_data_low[0], high_stride, low_stride, conf_threshold=0.3, original_image=True)
        print(box)
        draw_result_imshow(np.array(x_data[0]), box, coco_cls_names)
# #%%




#%%
import numpy as np
import cv2
import os


# display yolo result: each image with waitkey
def draw_result_imshow(image, annotation, class_names=None):
    disp_image = draw_result(image, annotation, class_names)
    cv2.imshow('image', disp_image)
    cv2.waitKey(0)



# display yolo result
def draw_result(image, annotation, class_names=None):

    # draw boxes
    img_size_y = image.shape[0]
    img_size_x = image.shape[1]

    # draw image first
    disp_image = image.copy()*255
    disp_image = disp_image.astype(np.uint8)
    
    # print(disp_image)
    for box in annotation:
        class_id, x, y, w, h = box
        x = x * img_size_x
        y = y * img_size_y
        w = w * img_size_x
        h = h * img_size_y
        cv2.rectangle(disp_image, (int(x-w/2), int(y-h/2)), (int(x+w/2), int(y+h/2)), (0, 255, 0), 2)


        # if coco dataset, display class name
        if class_names is not None:
            cls_name = class_names[int(class_id)]
        else: 
            cls_name = str(int(class_id))
            
        # display class name
        cv2.putText(disp_image, cls_name, (int(x-w/2+5), int(y+h/2-5)), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (255, 255, 255), 2)

    return disp_image





if __name__ == '__main__':
    data_path = 'C:/Users/osy04/Desktop/wok_me/project/yolo/images/train2017/'
    label_path = 'C:/Users/osy04/Desktop/wok_me/project/yolo/labels_bbox/train2017/'

    # coco dataset cls names
    coco_cls_names = ['person', 'bicycle', 'car', 'motorcycle', 'airplane', 'bus', 'train', 'truck', 'boat',
                    'traffic light', 'fire hydrant', 'stop sign', 'parking meter', 'bench', 'bird', 'cat', 'dog',
                    'horse', 'sheep', 'cow', 'elephant', 'bear', 'zebra', 'giraffe', 'backpack', 'umbrella',
                    'handbag', 'tie', 'suitcase', 'frisbee', 'skis', 'snowboard', 'sports ball', 'kite', 'baseball bat',
                    'baseball glove', 'skateboard', 'surfboard', 'tennis racket', 'bottle', 'wine glass', 'cup', 'fork',
                    'knife', 'spoon', 'bowl', 'banana', 'apple', 'sandwich', 'orange', 'broccoli', 'carrot', 'hot dog',
                    'pizza', 'donut', 'cake', 'chair', 'couch', 'potted plant', 'bed', 'dining table', 'toilet', 'tv',
                    'laptop', 'mouse', 'remote', 'keyboard', 'cell phone', 'microwave', 'oven', 'toaster', 'sink',
                    'refrigerator', 'book', 'clock', 'vase', 'scissors', 'teddy bear', 'hair drier', 'toothbrush']

    file_img = os.listdir(data_path)
    file_txt = os.listdir(label_path)

    for i in range(5):
        image = cv2.imread(data_path + file_img[i])/255.0
        image = cv2.resize(image, (416, 416))
        annotation = np.loadtxt(label_path + file_txt[i])

        if len(annotation.shape) == 1:
            annotation = annotation[np.newaxis, :]

        # print(file_img[i], annotation)
        print(image)
        # display result
        result = draw_result(image, annotation, coco_cls_names)
        cv2.imshow('image', result)
        cv2.waitKey(1000)





import tensorflow as tf
from tensorflow.keras import layers, Model

# ------------------------------------------------------------------
# Utility Blocks
# ------------------------------------------------------------------

def ConvBlock(x, filters, kernel_size=3, strides=1):
    """
    Convolution -> BatchNorm -> Activation (SiLU)
    """
    x = layers.Conv2D(filters, kernel_size, strides=strides, padding='same', use_bias=False)(x)
    x = layers.BatchNormalization()(x)
    x = layers.Activation(tf.nn.swish)(x)  # SiLU (swish) activation
    return x

def Focus(x, filters, kernel_size=3, strides=1):
    """
    Optimized Focus module using tf.nn.space_to_depth.
    This operator rearranges blocks of spatial data into depth.
    """
    # Replace manual slicing with the highly optimized space_to_depth op.
    x = tf.nn.space_to_depth(x, block_size=2)  # equivalent to slicing and concatenating four parts.
    x = ConvBlock(x, filters, kernel_size, strides)
    return x

def Bottleneck(x, filters, shortcut=True):
    """
    A simple bottleneck block.
    """
    y = ConvBlock(x, filters, kernel_size=1)
    y = ConvBlock(y, filters, kernel_size=3)
    if shortcut and x.shape[-1] == filters:
        y = layers.Add()([x, y])
    return y

def CSPBlock(x, filters, n=1):
    """
    A simple CSP (Cross Stage Partial) block.
    Splits the input into two parts, processes one part with bottleneck(s),
    then concatenates with the other part and fuses them.
    """
    route = ConvBlock(x, filters, kernel_size=1)
    x = ConvBlock(x, filters, kernel_size=1)
    for _ in range(n):
        x = Bottleneck(x, filters, shortcut=True)
    x = layers.Concatenate()([x, route])
    x = ConvBlock(x, filters, kernel_size=1)
    return x

def SPP(x, filters, pool_sizes=(5, 9, 13)):
    """
    Spatial Pyramid Pooling (SPP) block.
    Pools the features at different scales and concatenates them.
    """
    x = ConvBlock(x, filters // 2, kernel_size=1)
    pooled_outputs = [x]
    for pool_size in pool_sizes:
        pooled = layers.MaxPooling2D(pool_size, strides=1, padding='same')(x)
        pooled_outputs.append(pooled)
    x = layers.Concatenate()(pooled_outputs)
    x = ConvBlock(x, filters, kernel_size=1)
    return x

# ------------------------------------------------------------------
# Backbone (only two scales: small & large)
# ------------------------------------------------------------------

def build_backbone(inputs):
    """
    Backbone network (YOLOv8-style) that produces two scales:
      - feature_small: High-resolution features (for small object detection)
      - feature_large: Low-resolution features (for large object detection)
    """
    # Focus: reduce spatial dims and enrich channels.
    # x = Focus(inputs, filters=32, kernel_size=3) 
    x = layers.Conv2D(filters=32, kernel_size=3, strides=2, padding='same', activation='relu')(inputs)  # 360→180
    # Stage 1: High-resolution features.
    x = ConvBlock(x, 32, kernel_size=3, strides=2)  # 160 -> 80
    x = ConvBlock(x, 32, kernel_size=3, strides=2)  # 80 -> 40
    x = ConvBlock(x, 32, kernel_size=3, strides=2)  # 40 -> 20
    x = CSPBlock(x, filters=64, n=2)
    feature_small = x  # resolution ~160x160

    # Stage 2: Downsample to obtain coarse features.
    x = ConvBlock(x, 32, kernel_size=3, strides=2)  # 20 -> 10
    x = ConvBlock(x, 32, kernel_size=3, strides=2)  # 10 -> 5
    x = CSPBlock(x, filters=64, n=2)
    x = SPP(x, filters=64)
    feature_large = x  # resolution ~40x40

    return feature_small, feature_large

# ------------------------------------------------------------------
# Neck: Two-Scale Feature Fusion (Top-Down & Bottom-Up)
# ------------------------------------------------------------------

def build_neck_two_scales(feature_small, feature_large):
    """
    Fuses the high-resolution (feature_small) and low-resolution (feature_large) features.
    - Top-down: Upsample feature_large to fuse with feature_small.
    - Bottom-up: Downsample the fused feature to refine the coarse feature.
    Returns two refined features for detection.
    """
    # Top-down pathway: upsample feature_large and fuse with feature_small.
    up_feature = ConvBlock(feature_large, 64, kernel_size=1)
    up_feature = layers.UpSampling2D(size=4, interpolation='nearest')(up_feature) # 5 -> 20
    p_small = layers.Concatenate()([feature_small, up_feature])
    p_small = CSPBlock(p_small, filters=128, n=2)

    # Bottom-up pathway: downsample p_small back to coarse resolution.
    x = ConvBlock(p_small, 64, kernel_size=3, strides=2)  # 20 -> 10
    down_feature = ConvBlock(x, 64, kernel_size=3, strides=2)  # 10 -> 5
    p_large = layers.Concatenate()([feature_large, down_feature])
    p_large = CSPBlock(p_large, filters=128, n=2)

    return p_small, p_large

# ------------------------------------------------------------------
# YOLO Head
# ------------------------------------------------------------------

def build_yolo_head(x, num_classes):
    """
    YOLO head that predicts bounding box coordinates, objectness score,
    and class probabilities at each grid cell.
    The number of output channels = 4 (bbox) + 1 (objectness) + num_classes.
    """
    output_filters = 4 + 1 + num_classes
    x = layers.Conv2D(output_filters, kernel_size=1, strides=1, padding='same')(x)
    return x

# ------------------------------------------------------------------
# Assemble YOLOv8-Style Model (Two-Scale Version)
# ------------------------------------------------------------------

def create_yolov8_model(num_classes, image_size=640):
    """
    Create a YOLOv8-style model with two detection heads corresponding to:
      - A high-resolution feature (for small objects)
      - A low-resolution feature (for large objects)
    """
    inputs = tf.keras.Input(shape=(image_size, image_size, 3))
    
    # Backbone produces two scales.
    feature_small, feature_large = build_backbone(inputs)
    
    # Neck fuses the two scales.
    p_small, p_large = build_neck_two_scales(feature_small, feature_large)
    
    # Detection heads.
    head_small = build_yolo_head(p_small, num_classes)
    head_large = build_yolo_head(p_large, num_classes)
    
    model = Model(inputs, [head_small, head_large])
    return model

# Example: create the model for 80 classes (e.g., COCO) with 640x640 input images.
if __name__ == '__main__':
    num_classes = 80
    model = create_yolov8_model(num_classes, image_size=320)
    model.summary()

    # save model
    model.save('yolo_model8.h5')


#%%
import os
import tensorflow as tf
from datetime import datetime

# import modules
from . import yoloModel8
from . import createDataGenerator
from .yoloLoss import yolo_loss_tf


def fit_yolo_model(train_data_path, train_label_path, val_data_path, val_label_path, epochs=100, batch_size=64):

    # create folder name with current time to save models
    folder_name = datetime.now().strftime("%Y%m%d_%H%M")
    if not os.path.exists(folder_name):
        os.makedirs(folder_name)

    # model file name
    model_file_best = folder_name + '/yolo_model_best.keras'
    model_file_last = folder_name + '/yolo_model_last.keras'


    num_classes = 80  # Example number of classes
    img_size = 320

    # create model
    model = yoloModel8.create_yolov8_model(num_classes, img_size)
    high_stride = model.output_shape[0][1]
    low_stride = model.output_shape[1][1]
    # model.summary()

    # compile and fit model using adam solver with learning rate scheduler
    model.compile(optimizer=tf.keras.optimizers.Adam(learning_rate=0.002), loss=[yolo_loss_tf, yolo_loss_tf])

    # create data generator
    train_gen = createDataGenerator.create_tf_dataset(num_classes, high_stride, low_stride, img_size, train_data_path, train_label_path, batch_size=batch_size, shuffle=True, prob=0.5)
    val_gen = createDataGenerator.create_tf_dataset(num_classes, high_stride, low_stride, img_size, val_data_path, val_label_path, batch_size=batch_size, shuffle=False, prob=0.0)



    # create call back to save the best model
    model_checkpoint_callback = tf.keras.callbacks.ModelCheckpoint(
        filepath=model_file_best,
        save_weights_only=False,
        monitor='val_loss',
        mode='min',
        save_best_only=True)

    # save model every epoch with epoch number in file name
    model_checkpoint_callback2 = tf.keras.callbacks.ModelCheckpoint(
        filepath=folder_name + '/yolo_model_epoch_{epoch}.keras',
        save_weights_only=False,
        monitor='val_loss',
        mode='min',
        save_best_only=False)


    ## tensorboard callback
    # tensorboard_callback = TensorBoard(log_dir = './logs/' + datetime.now().strftime("%Y%m%d-%H%M%S"),
    #                                    histogram_freq=1, profile_batch='5,10')

    # train using generator
    model.fit(
        train_gen,
        # steps_per_epoch=4000,
        validation_data=val_gen,
        epochs=epochs,
        batch_size=batch_size,
        # use_multiprocessing=True,  # Use multiple processes to run the generator in parallel
        # workers=16,  # Number of worker processes
        callbacks=[model_checkpoint_callback, model_checkpoint_callback2]
    )

    # save the last model
    model.save(model_file_last)


#%% example usage
if __name__ == '__main__':
    #% encode data
    # encode training data
    train_data_path = 'C:/Users/somethingsomething'
    train_label_path = 'C:/Users/somethingsomething'

    # # encode validation data
    val_data_path = 'C:/Users/somethingsomething'
    val_label_path = 'C:/Users/somethingsomething'

    # fit model
    fit_yolo_model(train_data_path, train_label_path, val_data_path, val_label_path, epochs=10)





from oneclickai.YOLO import load_model, predict, draw_result
import os 
import cv2
import numpy as np

data_path = 'C:/Users/osy04/Desktop/wok_me/project/yolo/images/train2017/'
file_img = os.listdir(data_path)

# Load the model
model = load_model("YOLO_coco")

# Single image prediction
for i in range(5):
    image = cv2.imread(data_path + file_img[i])/255.0
    result_annotation = predict(model, image)
    result_image = draw_result(np.array(image), result_annotation)
    cv2.imshow('image', result_image)
    cv2.waitKey(0)


from .core import welcome_message

# yolo 모듈 통째로 import
from .YOLO import predict, predict_and_show, stream, load_model, draw_result, fit_yolo_model

# yolo 네임스페이스를 설정하기 위해 import
import oneclickai.YOLO

# 외부에서 import 시 노출될 함수 및 모듈 정의
__all__ = [
    'welcome_message',
    'add_numbers',
    'predict',
    'predict_and_show',
    'stream',
    'load_model',
    'draw_result',
    'fit_yolo_model',
]




#%%
import tensorflow as tf


def yolo_loss_tf(target, pred):

    # Loss weights
    lambda_coord = 5.0
    lambda_noobj = 0.5

    # -----------------------------
    # Separate out predictions
    # -----------------------------
    pred_x = pred[..., 0]
    pred_y = pred[..., 1]
    raw_w = pred[..., 2]
    raw_h = pred[..., 3]
    pred_w = tf.exp(raw_w)  # ensures width > 0
    pred_h = tf.exp(raw_h)  # ensures height > 0
    pred_conf = pred[..., 4]
    pred_logits = pred[..., 5:]  # raw logits for classes

    # -----------------------------
    # Separate out targets
    # -----------------------------
    tgt_x = target[..., 0]
    tgt_y = target[..., 1]
    tgt_w = target[..., 2]
    tgt_h = target[..., 3]
    tgt_conf = target[..., 4]      # 1 if object present, 0 otherwise
    tgt_classes = target[..., 5:]    # one-hot encoded classes

    # -----------------------------
    # Create object/no-object masks
    # -----------------------------
    obj_mask = tgt_conf           # mask for cells with objects
    noobj_mask = 1.0 - tgt_conf   # mask for cells without objects

    # -----------------------------
    # 1) Localization Loss (x, y, w, h)
    # -----------------------------
    # Loss for (x, y) coordinates using sum-of-squares (applied only where objects exist)
    loss_xy = obj_mask * ((tgt_x - pred_x) ** 2 + (tgt_y - pred_y) ** 2)
    loss_xy = tf.reduce_sum(loss_xy)

    # Loss for (w, h): using the squared difference of square roots
    eps = 1e-6  # small constant to avoid sqrt(0)
    loss_wh = obj_mask * (
        (tf.sqrt(tgt_w + eps) - tf.sqrt(pred_w + eps)) ** 2 +
        (tf.sqrt(tgt_h + eps) - tf.sqrt(pred_h + eps)) ** 2
    )
    loss_wh = tf.reduce_sum(loss_wh)

    loss_coord = lambda_coord * (loss_xy + loss_wh)

    # -----------------------------
    # 2) Objectness (Confidence) Loss
    # -----------------------------
    # Mean Squared Error (MSE) for confidence scores:
    loss_conf_obj = tf.reduce_sum(obj_mask * (tgt_conf - pred_conf) ** 2)
    loss_conf_noobj = tf.reduce_sum(noobj_mask * (tgt_conf - pred_conf) ** 2)
    loss_conf = loss_conf_obj + lambda_noobj * loss_conf_noobj

    # -----------------------------
    # 3) Classification Loss
    # -----------------------------
    # Compute binary cross-entropy loss (per class) for cells with objects
    bce = tf.nn.sigmoid_cross_entropy_with_logits(labels=tgt_classes, logits=pred_logits)
    bce = tf.reduce_sum(bce, axis=-1)  # sum over class probabilities
    loss_class = tf.reduce_sum(bce * obj_mask)

    # -----------------------------
    # Total Loss
    # -----------------------------
    total_loss = loss_coord + loss_conf + loss_class

    # tf.print('lossBox:', loss_coord, 'lossCnf:', loss_conf, 'lossCls:', loss_class)
    return total_loss




if __name__ == "__main__":
    pass

# %%



#%%
import numpy as np


# yolo output to bbox
def decode(high_prediction, low_prediction, high_stride, low_stride, conf_threshold, original_image=False):

    boxes = []
    
    # Each tuple consists of (prediction_array, grid_stride)
    scales = [
        (high_prediction, high_stride),
        (low_prediction, low_stride)
    ]
    
    for pred, stride in scales:
        # Assume the confidence is at index 4.
        confs = pred[..., 4]
        valid_idxs = np.where(confs >= conf_threshold)
        
        # Determine how many class scores are provided
        num_classes = pred.shape[2] - 5
        
        # Iterate over grid cells with high confidence
        for row, col in zip(*valid_idxs):
            # Unpack the parameters for this grid cell.
            # Expected layout: [tx, ty, tw, th, conf, p1, p2, ...]
            tx, ty, tw, th, conf, *class_scores = pred[row, col, :]
            
            # Decode the center coordinates relative to the whole image.
            # (tx, ty) are the offsets inside the grid cell.
            bx = (col + tx) / stride
            by = (row + ty) / stride
            
            # Decode the box dimensions.
            # If not using the "original" dimensions, assume tw and th are in log-space.
            if original_image:
                bw = tw
                bh = th
            else:
                bw = np.exp(tw)
                bh = np.exp(th)
            
            
            # Determine the predicted class by taking the argmax of the class scores.
            # If there are no class scores, default to -1.
            if num_classes > 0:
                predicted_class = int(np.argmax(class_scores))
            else:
                predicted_class = -1
            
            # Append the decoded box as integers.
            boxes.append([predicted_class,
                          bx,
                          by,
                          bw,
                          bh])
    return boxes


#%%
# ===========================
# Example Usage
# ===========================
if __name__ == "__main__":
    import yoloModel
    import createDataGenerator
    from drawImage import draw_result_imshow

    num_classes = 80  # Example number of classes
    img_size = 416
    data_path = 'C:/Users/osy04/Desktop/wok_me/project/yolo/images/val2017'
    label_path = 'C:/Users/osy04/Desktop/wok_me/project/yolo/labels_bbox/val2017'

    model = yoloModel.create_yolo_model(num_classes, img_size)
    high_stride = model.output_shape[0][1]
    low_stride = model.output_shape[1][1]
    model.summary()

    train_gen = createDataGenerator.data_generator(num_classes, high_stride, low_stride, img_size,data_path, label_path, batch_size=1, shuffle=True, prob=0.5)
    x_data, [y_data_high, y_data_low] = next(train_gen)

    # Decode predictions
    decoded_boxes = decode(y_data_high[0], y_data_low[0], high_stride,low_stride,conf_threshold=0.3,original_image=True)
    
    print("Decoded Boxes:")
    for box in decoded_boxes:
        print(box)

    draw_result_imshow(np.array(x_data[0]), decoded_boxes)
    